#!/bin/bash

export PATH=/bin:/sbin:/usr/bin:/usr/sbin:/usr/X11R6/bin

PV=
VG=
LV=

MODPROBE=
LVCREATE=
LVREMOVE=
LVRESIZE=
LVRENAME=

NOC="\e[0m"
RED="\e[1;31m"
GREEN="\e[1;32m"
YELLOW="\e[1;33m"

PC=
NC=0
MAX=0
NEW=1
RW='rw'
CHUNKSIZE=

PV_SIZE=0
VG_SIZE=0
LV_SIZE=0

PV_FREEg=0
PV_FREEb=0
VG_FREEg=0
VG_FREEb=0

SS_ARRAY=()
FRACTIONAL=0
VG_PREFIX='/dev'
SCRIPT=`basename $0`
TAB_CHAR=`echo -e '\011'`
COMMON_ARGS='--noheadings --nosuffix'

PVS=`which pvs 2> /dev/null`
VGS=`which vgs 2> /dev/null`
LVS=`which lvs 2> /dev/null`
MODPROBE=`which modprobe 2> /dev/null`
LVCREATE=`which lvcreate 2> /dev/null`
LVREMOVE=`which lvremove 2> /dev/null`
LVRESIZE=`which lvresize 2> /dev/null`
LVRENAME=`which lvrename 2> /dev/null`

print_usage () {
   echo "Usage: $SCRIPT {-vg <volume group> -lv <logical volume> || -dv </dev/vg/lv>} -pc <initial % of LV for snapshot> [-ch <size>] [-lt <limit>] [-ro] [-pv] [-nc]"

   echo -e "\n-ro = Read-only"
   echo "-nc = Run without confirmation"
   echo -e "-pv = Create SS on same physical volume as the origin\n"

   echo "When supplying a limit (maximum size of free VG space to use) supply a suffix of either b, k, m, g or t."
}

trim () {
   sed "s/^[ $TAB_CHAR]\{1,\}//g;s/[ $TAB_CHAR]\{1,\}$//g" < /dev/stdin
}

n_percent () { # Argument 1 = total, 2 = percentage, [Optional 3 = precision]
   local precision=${3:-0} # Default is no decimal point

   # BC math functions taken from:
   # http://www.cyreksoft.yorks.com/gnu-bc/funcs.bc

   if [ "x$1" != "x" ] && [ "x$2" != "x" ]
   then
      echo "scale=$precision;

      # Round to next integer nearest 0:  -1.99 -> 1, 0.99 -> 0
      define int(x) { auto os;os=scale;scale=0;x/=1;scale=os;return(x) }

      # Round down to integer below x
      define floor(x) { auto xx;xx=int(x);if(xx>x)xx-=1;return(xx) }

      # Round up to integer above x
      define ceil(x) { return(-floor(-x)) }

      ceil (($1 / 100) * $2)" | bc -l
   fi
}

percent_n () { # Argument 1 = part, 2 = total, [Optional 3 = precision]
   local precision=${3:-0} # Default is no decimal point

   if [ "x$1" != "x" ] && [ "x$2" != "x" ]
   then
      echo "scale=$precision; ($1 / $2) * 100" | bc -l
   fi
}

load_kernel_module () { # [Optional argument 1 = kmod]
   local module=${1:-dm-snapshot}

   $MODPROBE $module > /dev/null 2>&1
   return $?
}

lt () { # Argument 1 = x, 2 = n, [Optional 3 = precision]
   local precision=${3:-0} # Default is no decimal point
   echo "scale=$precision; $1 < $2" | bc
}

gt () { # Argument 1 = x, 2 = n, [Optional 3 = precision]
   local precision=${3:-0} # Default is no decimal point
   echo "scale=$precision; $1 > $2" | bc
}

set_max () {
   local free=`[ "x$PV" != "x" ] && echo $PV_FREEb || echo $VG_FREEb`

   if [ "$MAX" = "0" ]
   then
      MAX=$free
   else
      local len=$((`echo $MAX | wc -c` - 1))
      local suf="`echo $MAX | cut -b $len`"
      MAX=${MAX%*$suf}

      case $suf in
         b|B)
            ;;
         k|K)
            MAX=`echo "scale=$FRACTIONAL;$MAX * 1024" | bc`
            ;;
         m|M)
            MAX=`echo "scale=$FRACTIONAL;$MAX * 1024 * 1024" | bc`
            ;;
         g|G)
            MAX=`echo "scale=$FRACTIONAL;$MAX * 1024 * 1024 * 1024" | bc`
            ;;
         t|T)
            MAX=`echo "scale=$FRACTIONAL;$MAX * 1024 * 1024 * 1024 * 1024" | bc`
            ;;
         *)
            ;;
      esac
   fi

   if [ `gt $MAX $free $FRACTIONAL` -eq 1 ]
   then
      echo "Your chosen limit (${MAX}b) is greater than the amount of available space."
      exit 1
   fi
}

list_vgs () {
   local arg='-o vg_name'
   $VGS $COMMON_ARGS $arg | trim
}

list_pvs () { # [Optional argument = vg]
   local cmd=
   local arg='-o pv_name'

   if [ "x$1" != "x" ]
   then
      arg="--separator , $arg,vg_name"
      cmd="grep ,$1$"
   else
      cmd="cat"
   fi

   $PVS $COMMON_ARGS $arg | trim | $cmd
}

list_lvs () { # [Optional argument = vg]
   local cmd=
   local arg='-o lv_name'

   if [ "x$1" != "x" ]
   then
      arg="--separator , $arg,vg_name"
      cmd="grep ,$1$"
   else
      cmd="cat"
   fi

   $LVS $COMMON_ARGS $arg | trim | $cmd
}

test_pv () { # Argument 1 = pv
   local arg='-o pv_name'
   $PVS $COMMON_ARGS $arg | trim | grep -c ^$1$
}

test_vg () { # Argument 1 = vg
   local arg='-o vg_name'
   $VGS $COMMON_ARGS $arg | trim | grep -c ^$1$
}

test_lv () { # Argument 1 = lv
   local arg='-o lv_name'
   $LVS $COMMON_ARGS $arg | trim | grep -c ^$1$
}

get_size () { # Argument 1 = type (pv, vg, lv), 2 = object name, [Optional 3 = suffix]
   if [ $# -lt 2 ]
   then
      return 1
   fi

   local suffix=${3:-b} # Default is bytes
   local arg="--units 1${suffix} -o "

   (
      case $1 in
         vg)
            $VGS $COMMON_ARGS $arg vg_size $2
            ;;
         pv)
            $PVS $COMMON_ARGS $arg pv_size $2
            ;;
         lv)
            $LVS $COMMON_ARGS $arg lv_size $2
            ;;
         *)
            return 1
            ;;
      esac
   ) | trim
}

get_free () { # Argument 1 = type (pv, vg, lv), 2 = object name, [Optional 3 = suffix]
   if [ $# -lt 2 ]
   then
      return 1
   fi

   local suffix=${3:-b} # Default is bytes
   local arg="--units 1${suffix} -o "

   (
      case $1 in
         vg)
            $VGS $COMMON_ARGS $arg vg_free $2
            ;;
         pv)
            $PVS $COMMON_ARGS $arg pv_free $2
            ;;
         lv)
            $LVS $COMMON_ARGS $arg lv_free $2
            ;;
         *)
            return 1
            ;;
      esac
   ) | trim
}

percent_sss () {
   local arg='-o snap_percent'

   if [ "x$1" = "x" ]
   then
      return 1
   fi

   $LVS $COMMON_ARGS $arg $VG_PREFIX/$VG/$1 | trim
}

determine_sss () { # Argument 1 = %, 2 = vg, 3 = lv
   local lvbs=`get_size lv $VG_PREFIX/$2/$3 b` # LV size in bytes
   local pct=`n_percent $lvbs $1 $FRACTIONAL` # % of lvbs
   local -a index=( $pct )
   local total=0
   local j=$pct
   local i=0

   if [ `lt $j $MAX` -ne 1 ]
   then
      echo "${MAX}b of '$VG' does not allow even 1 $PC% SS of '$LV'."
      return 1
   fi

   echo "Snapshot segment break-down:"

   for ((i=0 ; `lt $total $MAX` ; i=i+1))
   do
      if [ $i -gt 0 ]
      then
         j=$((${index[$(($i - 1))]} + $pct))
         index=( ${index[@]} $j)
      fi

      total=$(echo "scale=0;`echo ${index[@]} | tr ' ' '+'`" | bc)
   done

   unset index[$(($i - 1))]
   SS_ARRAY=( ${index[@]} )

   return 0
}

confirm_sss () {
   local freeb=`[ "x$PV" != "x" ] && echo $PV_FREEb || echo $VG_FREEb`
   local freeg=`[ "x$PV" != "x" ] && echo $PV_FREEg || echo $VG_FREEg`

   echo

   for i in ${!SS_ARRAY[@]}
   do
      echo -e "\tSS ${YELLOW}$i${NOC}: ${SS_ARRAY[$i]}b"
   done

   i=$(echo "scale=0;`echo ${SS_ARRAY[@]} | tr ' ' '+'`" | bc)
   p=`echo "scale=2;(($freeb - $i) / $freeb) * 100" | bc`
   echo -e "\n\tRemainder: $(($freeb - $i))b (${p}% of ${freeg}G)\n"

   if [ "$NC" = "1" ] # No confirmation has been requested
   then
      return 0
   fi

   read -n 1 -e -p "Go ahead with this setup? " confirm

   case $confirm in
      y|Y)
         return 0
         ;;
      n|N)
         exit 1
         ;;
      *)
         confirm_sss
         ;;
   esac
}

get_mp () {
   grep "^$1 " /proc/mounts | awk '{print $2}'
}

get_fs () {
   grep "^$1 " /proc/mounts | awk '{print $3}'
}

get_lvs_pvs () {
   local cmd="grep ^$2,$1,\/"
   local arg='--separator , -o lv_name,vg_name,devices'
   local lv_pv=`$LVS $COMMON_ARGS $arg | trim | $cmd | awk -F , '{print $3}'`

   echo ${lv_pv%%(*)}
}

create_sss () {
   echo -e "\nCreating snapshot LVs:\n"

   for i in ${!SS_ARRAY[@]}
   do
      local name="$LV-SS-$i"
      local size=$((${SS_ARRAY[$i]} /  1024))

      echo -n -e "\t${YELLOW}${name}${NOC}..."
      $LVCREATE $CHUNKSIZE -A y -n ${name} -L ${size}k -s -p ${RW} $VG_PREFIX/$VG/$LV $PV > /dev/null 2>&1

      if [ $? -eq 0 ]
      then
         echo -e "${GREEN}success${NOC}."
      else
         echo -e "${RED}failure${NOC}."
      fi
   done

   echo
}

test_suite () {
   echo -e "\nTest list_pvs(), list_lvs(), get_size() with arguments:\n"

   for vg in `list_vgs`
   do
      echo "VG $vg contains:"

      for pv in `list_pvs $vg | cut -d , -f 1`
      do
         echo -e "\tPV $pv is `get_size pv $pv`b (`get_size pv $pv g`G)"
      done

      for lv in `list_lvs $vg | cut -d , -f 1`
      do
         echo -e "\tLV $lv is `get_size lv $VG_PREFIX/$vg/$lv`b (`get_size lv $VG_PREFIX/$vg/$lv g`G)"
      done

      echo -e "VG $vg total capacity is `get_size vg $vg`b (`get_size vg $vg g`G)\n"
   done
}

gather_sss () { # [Optional argument = print sizes]
   local -a sss=( `list_lvs $VG | cut -d , -f 1 | grep "^${LV}-SS-[0-99]" | sort -g -t - -k 3 | tr '\n' ' '` )

   if [ ${#sss[@]} -eq 0 ]
   then
      return 1
   fi

   SS_ARRAY=( ${sss[@]} )

   if [ "x$1" != "x" ]
   then
      echo -e "Discovered snapshot LVs:\n"

      for lv in ${SS_ARRAY[@]}
      do
         echo -e "\t${YELLOW}$lv${NOC} (`get_size lv $VG_PREFIX/$VG/$lv g`G, `percent_sss $lv`% full)"
      done

      echo
   fi

   return 0
}

rotate_sss () {
   local lvbs=`get_size lv $VG_PREFIX/$2/$3 b` # LV size in bytes
   local pct=`n_percent $lvbs $1 $FRACTIONAL` # % of lvbs
   local -a sizes=( )

   for i in ${!SS_ARRAY[@]}
   do
      sizes=( ${sizes[@]} `get_size lv $VG_PREFIX/$VG/${SS_ARRAY[$i]} b` )
   done

   if [ `echo ${sizes[@]} | tr ' ' '\n' | sort -g -r | head -n 1` -ne ${sizes[$i]} ]
   then
      read -n 1 -e -p "Seems ${SS_ARRAY[$i]} is not the oldest SS, continue anyway? " confirm

      case $confirm in
         y|Y)
            ;;
         n|N)
            exit 1
            ;;
         *)
            rotate_sss $1 $2 $3
            ;;
      esac
   fi

   echo -e "Running 'rotate' sequence on $LV SS:\n"

   echo -ne "\t${RED}Removing${NOC} 'old' SS ${SS_ARRAY[$i]}..."
   $LVREMOVE -f $VG_PREFIX/$VG/${SS_ARRAY[$i]} > /dev/null 2>&1

   if [ $? -eq 0 ]
   then
      echo -e "${GREEN}success${NOC}."
      unset SS_ARRAY[$i]
   else
      echo -e "${RED}failure${NOC}."
      exit 1
   fi

   for i in `seq 0 $((${#SS_ARRAY[@]} - 1)) | sort -g -r`
   do
      local old=${SS_ARRAY[$i]}
      local new="$LV-SS-$(($i + 1))"
      local size=$((${sizes[$(($i + 1))]} / 1024))

      echo -ne "\t${RED}Resizing${NOC} $old from ${sizes[$i]}b to ${sizes[$(($i + 1))]}b..."
      $LVRESIZE -A y -n -L ${size}k $VG_PREFIX/$VG/$old $PV > /dev/null 2>&1

      if [ $? -eq 0 ]
      then
         echo -e "${GREEN}success${NOC}."
      else
         echo -e "${RED}failure${NOC}."
      fi

      echo -ne "\t${RED}Renaming${NOC} $old to ${new}..."
      $LVRENAME $VG_PREFIX/$VG/$old $VG_PREFIX/$VG/$new > /dev/null 2>&1

      if [ $? -eq 0 ]
      then
         echo -e "${GREEN}success${NOC}."
      else
         echo -e "${RED}failure${NOC}."
      fi
   done

   echo -ne "\t${RED}Recreating${NOC} ${LV}-SS-${i} (${sizes[0]}b)..."
   $LVCREATE $CHUNKSIZE -A y -n ${LV}-SS-${i} -L $((${sizes[0]} / 1024))k -s -p ${RW} $VG_PREFIX/$VG/$LV $PV > /dev/null 2>&1

   if [ $? -eq 0 ]
   then
      echo -e "${GREEN}success${NOC}."
   else
      echo -e "${RED}failure${NOC}."
   fi
}

init () {
   if [ "x$VG" = "x" ]
   then
      echo "No VG supplied or discovered from your '-dv' argument."
      exit 1
   fi

   if [ "x$LV" = "x" ]
   then
      echo "No LV supplied or discovered from your '-dv' argument."
      exit 1
   fi

   if [ "x$PC" = "x" ]
   then
      echo "No growth rate (percentage) given."
      exit 1
   fi

   if [ `test_vg $VG` -ne 1 ]
   then
      echo "VG '$VG' not found."
      exit 1
   fi

   if [ `test_lv $LV` -ne 1 ]
   then
      echo "LV '$LV' not found."
      exit 1
   fi

   if [ "$PV" = "1" ]
   then
      PV=`get_lvs_pvs $VG $LV`
      PV_SIZE=`get_size pv $PV g`
      PV_FREEb=`get_free pv $PV b`
      PV_FREEg=`get_free pv $PV g`
   else
      PV=
   fi

   VG_SIZE=`get_size vg $VG g`
   LV_SIZE=`get_size lv $VG_PREFIX/$VG/$LV g`

   VG_FREEb=`get_free vg $VG b`
   VG_FREEg=`get_free vg $VG g`

   [ "x$PV" != "x" ] && echo "Source PV: '$PV' (${PV_SIZE}G)"
   echo "Source VG: '$VG' (${VG_SIZE}G)"
   echo "Source LV: '$LV' (${LV_SIZE}G)"

   echo -e "\nChosen growth rate: $PC%"
   [ "x$PV" != "x" ] && echo "PV '$PV' Free space: ${PV_FREEb}b (${PV_FREEg}G)"
   echo -e "VG '$VG' Free space: ${VG_FREEb}b (${VG_FREEg}G)\n"
}

if [ `id -u` -ne 0 ]
then
   echo "You ain't got r00t!"
   exit 1
fi

for tool in $PVS $VGS $LVS $MODPROBE $LVCREATE $LVREMOVE $LVRESIZE $LVRENAME
do
   if [ "x$tool" = "x" ] || [ ! -x $tool ]
   then
      echo "Not all the LVM2 tools could be found."
      exit 1
   fi
done

load_kernel_module

if [ $# -lt 4 ]
then
   print_usage
   test_suite
   exit 1
fi

for arg in "$@"
do
   case $arg in
      "-vg")
         shift
         VG=$1
         ;;
      "-lv")
         shift
         LV=$1
         ;;
      "-dv")
         shift
         LV=${1##*/}
         VG=${1%/$LV}
         VG=${VG##$VG_PREFIX/}
         ;;
      "-pc") # Growth percentage
         shift
         PC=$1

         if [[ "$PC" =~ '\.' ]]
         then
            FRACTIONAL=${PC##*.}
            FRACTIONAL=${#FRACTIONAL}
         fi
         ;;
      "-lt")
         shift
         MAX=$1
         ;;
      "-ch")
         shift
         CHUNKSIZE="-c $1"
         ;;
      "-ro")
         shift
         RW='r'
         ;;
      "-pv")
         shift
         PV=1
         ;;
      "-nc")
         shift
         NC=1
         ;;
      *)
         shift
         ;;
   esac
done

init
set_max
gather_sss # sss = SnapShot Segment

if [ $? -ne 0 ]
then
   determine_sss $PC $VG $LV

   if [ $? -eq 0 ]
   then
      confirm_sss
      create_sss
   fi
else
   unset NEW
fi

gather_sss 1 # 1 indicates we wants sizes printed

if [ -z $NEW ]
then
   rotate_sss $PC $VG $LV
fi

exit 0
